/* CAP4601 Intro to AI      Final Project
 * Author: Rick Tilley      Date: 7/31/2013
 *
 * Description:
 *   Regression used to navigate in a game schenario.
 * There is a fly and a fly-killer.  They both are similar
 * in their design except one flee's and one chases.
 * Using gradient descent to predict opponent actions and
 * decrease likelyhood of failure.
 */
#include <iostream>
#include <cmath>
//#include <print.hpp>
#include <ctime>
#include <cstdlib>
#include <math.h>

#define FLIGHT_LOG_LENGTH 20
#define pi 3.1415926535897932384626433832795028841971693993751058209749445923078164

static const double DEGREE = 180. / pi;
static const double PI_2 = pi / 2.0;

class Fly {
	enum { T, X, N };
	struct FlightLog {
		double data[FLIGHT_LOG_LENGTH][2];
		int pasti;  
		bool looper;
		double theta;
		double xnot, ynot;
		FlightLog()
		{	pasti = -1;
			looper = false;
		}
	};
	struct Weights {
		double w[2];
		double theta;
		double xnot, ynot;
	};

	struct Stats {
		double xnot, ynot;
		double n, sumx, sumy, dx, dy, dsqx, dsqy, dxy;
		Stats() {
			xnot=0;ynot=0;n=0;sumx=0;sumy=0;dx=0;dy=0;dsqx=0;dxy=0;dsqy=0;
		}
	};

public:
	Fly() : learning_rate(0.1)
	{
		srand((unsigned)time(0));
		for (int w=0; w < 2; w++)
		{ weights.w[w] = ((rand() % 200) / 100.0) - 1.0;
		  weights.w[w] = ((rand() % 200) / 100.0) - 1.0;
		}
		weights.xnot = ((rand() % 100) / 10.0) - 5.0;
		weights.ynot = ((rand() % 100) / 10.0) - 5.0;
		weights.theta = (double)(rand() % 360) * pi / 180.0;
		abstime = 0;
		reltime = 0;
		predator = false;
		attention = 3;
		attention2 = 5;
		perception = 15;
		past.data[0][0] = weights.xnot;
		past.data[0][1] = weights.ynot;
	}
	void setAttack(bool p) { predator = p; }
	void setAttention(int a) { attention = a; }
	void setAttentionSpan(int a) { attention2 = a; }
	void pickpath(const FlightLog& log);
	int gettime() { return (int)floor(abstime); }
	void settime(int t) { abstime = t; }
	bool move1();
	void printpos();
	void printweights();  // for testing purposes
	const FlightLog& dumppath() { return past; }
	void setWeight(double w[N], double th, double x, double y)
	{  for (int i = 0; i < N; i++)
	     weights.w[i] = w[i];
	   weights.theta = th;
	   weights.xnot = x;  weights.ynot = y;
	}
	double distance(Fly& foe) const;		

protected:
	Weights weights;
	double abstime, reltime;
	FlightLog past;
	int attention, attention2;
	int perception;
	bool predator;
	double const learning_rate;


	void recalibrate(Weights &me);
	FlightLog rotateLog(FlightLog log);
	double predicty(double x, const Weights &w);
	FlightLog prepLog(FlightLog log);
	int screenpos(int j, double x, double y, double theta);
	FlightLog& regress(FlightLog& log, Weights &yourw);
};

void Fly::pickpath(const FlightLog& log)
{	Weights you;		// opponent
//	Weights me;		// actor
	double wrkh[2], dwrk[2];
	wrkh[0] = wrkh[1] = dwrk[0] = dwrk[1] = 0;
	double px,py;
	double distance;
	FlightLog predicts;
	FlightLog newtraj;
	FlightLog temp;
	// analyze opponent first
	for (int w=0; w < N; w++)
	  you.w[w] = (double)((rand() % 100) / 100.0) - 0.5;
//	temp = rotateLog(log);
	temp = prepLog(log);
	regress( temp, you);
	predicts.data[0][0] = past.data[past.pasti][0];
	predicts.data[0][1] = past.data[past.pasti][1];
	distance = (double) sqrt( pow(log.data[log.pasti][0] - predicts.data[0][0],2)+pow(log.data[log.pasti][1] - predicts.data[0][1],2) );
	// predict half speed
	distance = distance / 2.0;
	if (predator)
	{
	  px = log.data[log.pasti][0]+distance;
	}
	else
	{ // follows opponents tail in a way, effectively runs away
	  px = log.data[log.pasti][0]-distance;
	}
      py = predicty(px / cos(you.theta), you);
	  predicts.data[FLIGHT_LOG_LENGTH-1][0] = px;
	  predicts.data[FLIGHT_LOG_LENGTH-1][1] = py;
	predicts.pasti = FLIGHT_LOG_LENGTH-1;
	// add some noise inbetween start and destination, draw a line
	px = predicts.data[FLIGHT_LOG_LENGTH-1][0] - predicts.data[0][0];
	py = predicts.data[FLIGHT_LOG_LENGTH-1][1] - predicts.data[0][1];
	weights.w[1] = py/px;
	weights.w[0] = predicts.data[0][1];
	py = px/FLIGHT_LOG_LENGTH;
	// draw a wobbly straight line from me to (pseudo)you
	for (int i = 1; i < FLIGHT_LOG_LENGTH-1; i++)
	{
		predicts.data[i][0] = (double)i*py;
		predicts.data[i][1] = ((rand() % 100) / 100.0 - 0.5) + predicts.data[i][0]*weights.w[1]+weights.w[0];
		predicts.data[i][0] += predicts.data[0][0];
	}
//	newtraj = rotateLog(predicts);
	newtraj = prepLog(predicts);
	for (int w=0; w < N; w++)
	  weights.w[w] = (double)((rand() % 100) / 100.0) - 0.5;
	regress(newtraj, weights);
	// ensure weights are within limits, normalize
	recalibrate(weights);
}

Fly::FlightLog Fly::prepLog(FlightLog log)
{ 
  FlightLog temp;
	if (log.looper)
	{    int k = 0;
	   for (int i = log.pasti+1; i < FLIGHT_LOG_LENGTH; i++)
	   {  temp.data[k][0] = log.data[i][0];
	      temp.data[k++][1] = log.data[i][1];
	   }
	   for (int i = 0; i <= log.pasti; i++)
	   {  temp.data[k][0] = log.data[i][0];
	      temp.data[k++][1] = log.data[i][1];
	   }
	   temp.pasti = FLIGHT_LOG_LENGTH-1;
	}
	else
	{
	   for (int i = 0; i <= log.pasti; i++)
	   {  temp.data[i][0] = log.data[i][0];
	      temp.data[i][1] = log.data[i][1];
	   }
	   temp.pasti = log.pasti;
	}
	temp.theta = log.theta;
	temp.xnot = log.xnot;   
	temp.ynot = log.ynot;
  return temp;
}

/*
// deprecated, was used for gradient descent
Fly::FlightLog Fly::rotateLog(FlightLog log)
{ 
  double xnot, ynot;
  double normal, theta; 
  FlightLog temp, rotated;
	if (log.looper)
	{    int k = 0;
	   for (int i = log.pasti+1; i < FLIGHT_LOG_LENGTH; i++)
	   {  temp.data[k][0] = log.data[i][0];
	      temp.data[k++][1] = log.data[i][1];
	   }
	   for (int i = 0; i <= log.pasti; i++)
	   {  temp.data[k][0] = log.data[i][0];
	      temp.data[k++][1] = log.data[i][1];
	   }
	   temp.pasti = FLIGHT_LOG_LENGTH-1;
	}
	else
	{
	   for (int i = 0; i <= log.pasti; i++)
	   {  temp.data[i][0] = log.data[i][0];
	      temp.data[i][1] = log.data[i][1];
	   }
	   temp.pasti = log.pasti;
	}
	// rotate to slope of line to make y' = n*x' + y0;
	{
	  double ndx, ndy;
	  ndx = temp.data[temp.pasti][0] - temp.data[0][0];
	  ndy = temp.data[temp.pasti][1] - temp.data[0][1];
	  if (ndx == 0) ndx = 0.000001;
      normal = (ndy/ndx);
	  theta = atan(normal);  
	  ynot = temp.data[0][1];
	  xnot = temp.data[0][0];
	  for (int i = 0; i <= temp.pasti; i++)
	  {  ndx = temp.data[i][0]-xnot;  ndy = temp.data[i][1]-ynot;
		  rotated.data[i][0] = ndx*cos(theta)+ndy*sin(theta);
		  rotated.data[i][1] = -1.0*ndx*sin(theta)+ndy*cos(theta);
	  }
	}
	rotated.theta = theta;
	rotated.xnot = xnot;   
	rotated.ynot = ynot;
	rotated.pasti = temp.pasti;
  return rotated;
}
*/
void Fly::recalibrate(Weights &me)
{  // normalize/limit big weights so we limit total velocity
	double w1 = me.w[1];
	double max = 5.0;
	if ( (w1 > max) | (w1 < -1.0*max ))
	{
		if (w1 > max)
			me.w[1] = max;
		else
			me.w[1] = -1.0*max;
//		me.w[1] = sin(me.theta) * 2.0;
		me.w[0] = past.data[past.pasti][1] - me.w[1]*past.data[past.pasti][0];
	}
}

// uses solved formula for regression
Fly::FlightLog& Fly::regress(FlightLog& log, Weights &yourw)
{
	double x, y, n,dx,dy;
	Stats s;
	x = 0; y = 0; dx = 0; dy = 0;
	s.n = n = (double) log.pasti+1;
	for (int i = 0; i < log.pasti; i++)
	{
		x = x + log.data[i][0];
		y = y + log.data[i][1];
	}
	s.sumx = x;  s.sumy = y;
	s.xnot = x / n;
	s.ynot = y / n;
	// init to 0 by constructor
	for (int i = 0; i < log.pasti; i++)
	{
		dx = log.data[i][0] - s.xnot;
		dy = log.data[i][1] - s.ynot;
		s.dx += dx;  s.dy += dy;
		s.dsqx += dx*dx;
		s.dsqy += dy*dy;
		s.dxy += dx*dy;
	}
	//  double n, sumx, sumy, dx, dy, dsqx, dxy;
	// w1 = dsqry / dsqrx;
	// w0 = 0 i guess
//	yourw.w[1] = (s.n*s.dxy - s.dx*s.dy) / (s.n*s.dsqx - s.dx*s.dx);
	yourw.w[1] = s.dxy / s.dsqx;
//	yourw.w[1] = yourw.w[1] / (sqrt( s.dxy*s.dxy + s.dsqx*s.dsqx ) );
//	yourw.w[0] = (s.dy - yourw.w[1] * s.dx) / s.n;
	yourw.w[0] = -1 * yourw.w[1] * s.xnot - s.ynot;


	double angle_1 = ( 1./2. ) * atan2( 2. * s.dxy, s.dsqx - s.dsqy );
	double angle_2 = angle_1 + PI_2;
	if ( PI_2 < angle_2 ) {
		angle_2 = angle_1 - PI_2;
	}
	if ( 0 < ( 2. * cos( 2. * angle_1 ) * ( s.dsqx - s.dsqy ) + 4. * sin( 2. * angle_1 ) * s.dxy )  ) {
		yourw.theta = angle_1;
	} else {
		yourw.theta = angle_2;
	}
	yourw.xnot = s.xnot;
	yourw.ynot = s.ynot;
	return log;
}

/*
// uses gradient descent for regression
Fly::FlightLog& Fly::regress(FlightLog& log, Weights &yourw)
{	Weights you;
	double const threshold = 0.000001;
	double w_diff = threshold+4.0;
	double norm;
	size_t iters = 2;
	size_t const max_iterations = 1024;
	size_t datasize = sizeof(log) / sizeof(log.data[0][0]);
	double yyou, cost, xj;
	int endi = log.pasti+1;
   yourw.theta = log.theta;
   yourw.xnot = log.xnot;  yourw.ynot = log.ynot;
// gradient analysis
  if (endi > 1)
  {
	while (iters <= max_iterations && threshold < w_diff )
	{ // first backup variables for concurrency
	  for (int i = 0; i < N; i++)
	    you.w[i] = yourw.w[i];
	  for (int j = 0; j < N; j++)
	  { cost = 0;  norm = 0;
	    for (int k = 0; k < endi; k++)
	    { // calculate cost
		  yyou = yourw.w[0]+yourw.w[1]*log.data[k][0];
		  if (j == 0)
		    xj = 1;
		  else
		    xj = log.data[k][0];
		  cost = cost + (yyou - log.data[k][1])*xj;
		  norm += xj*xj;
	    }
	    // multiply by alpha and find new w[j]
//		cost = cost / (double)endi;  // biased normalization
		// gotta normalize somehow
		cost = cost / (double) endi;  // normalization
	    you.w[j] = you.w[j] - learning_rate*cost;
	  }
	  // calculate distance traveled
	  // to see if we are moving fast enough
	  w_diff = 0;
	  for (int j = 0; j < N; j++)
	    w_diff += pow(you.w[j] - yourw.w[j], 2);
	  w_diff = sqrt(w_diff);
	  iters++;
	  // update for concurrency
	  for (int i = 0; i < N; i++)
	  {  yourw.w[i] = you.w[i]; }
	}
  }
  return log;
}
*/
/*
// deprecated, used for rotation translation
int Fly::screenpos(int j, double x, double y, double theta)
{  int ret = 0;
   double mycos, mysin;
   mycos = cos(theta*-1.0);
   mysin = sin(theta*-1.0);
	if (j == 0)
	  ret = (int)( x*mycos+y*mysin );
	if (j == 1)
	  ret = (int)( y*mycos - x*mysin );
	return ret;
}
*/
double Fly::predicty(double x, const Weights &w)
{
	return (w.w[0]+w.w[1]*x);
}

bool Fly::move1()
{   ++abstime;
	++reltime;
	double x = 0,y;
	if (past.pasti != -1)
	  x = past.data[past.pasti][0];
	else
	  x = weights.xnot;  // first time

	if (++past.pasti >= FLIGHT_LOG_LENGTH)
	{
		past.pasti = 0;  past.looper = true;
	}
	if ( ( (weights.theta > PI_2) & (weights.theta < 3.0*PI_2) ) | 
	    ( (weights.theta < -1.0 * PI_2) & (weights.theta > -3.0 * PI_2) ) )
	  x -= 1.0;
	else
	  x += 1.0;
//	x += cos(weights.theta);
	y = predicty(x, weights);
	past.data[past.pasti][0] = x;
	past.data[past.pasti][1] = y;
//	past.data[past.pasti][0] = screenpos(0,x,y,weights.theta) + weights.xnot;
//	past.data[past.pasti][1] = screenpos(1,x,y,weights.theta) + weights.ynot;

	if (attention > 0)
	{	--attention;		return false;	}
	else
	{	attention = attention2;	return true; }
}

void Fly::printpos()
{
	std::cout << "{ " << past.data[past.pasti][0]
			<< ", " << past.data[past.pasti][1]
			<< " }";
}

void Fly::printweights()
{  
	std::cout << "{ [w0:" << weights.w[0] << "::w1:" << weights.w[1]	
	<< "::deg:" << weights.theta*DEGREE << " or " << weights.theta << " ] }";
/*
	std::cout << "{ [ " << weights.w[0] << ", " << weights.w[1]	
	<< ", " << alpha << weights.theta << ", +(" << weights.xnot << "," << weights.ynot << ")"
	<< " ] }";
	*/
}

double Fly::distance(Fly& foe) const
{
	return sqrt( pow( foe.past.data[foe.past.pasti][0] - past.data[past.pasti][0], 2)
		       + pow( foe.past.data[foe.past.pasti][1] - past.data[past.pasti][1], 2) );
}
